/* -*- Mode: C -*-
 * Stream-based lexical analyzer.  
 * 30 Jan 85.  
 * Ben Moszkowski
 * Updated 12 Feb 85; 19 Apr 85
 *
 * The process lexer_process(Signal,Char_receiver,Token_stream) takes the
 * character input from the port Char_receiver and lexes it, outputting
 * tokens on the port Token_sender.
 *
 */

/* Uses stream routines. */
/* load(streams). */

/* Lexer process */
define lexer_process(Signal,Char_receiver,Token_sender) = process {
    {
	begin_process(Signal)
	and stable_ports()
    };
    {
	lex_expr() and stable Signal
    };
    {
	finish_process(Signal)
	and stable_ports()
    }
} where


/* Body of lexer_process */
define lex_expr() = {
    open_char_and_token_ports();
    while not end_of_chars() do lex_one_token();
    close_char_and_token_ports()	
} and


/*  Lex a single token */
define lex_one_token() = {
    if is_digit(char_data()) then lex_number()
    else if char_data()=" " then lex_spaces()
    else lex_operator()
} and


/* Lex a number */
define lex_number() = exists Num:{
    Num=0 and
    while not end_of_chars() and is_digit(char_data()) do {
	{
	    skip
	    and Num:= 10*Num+ascii(char_data())-ascii("0")
	    and stable_ports()
	};
	{
	    get_char() and stable Num
	}
    };
    {
	put_token(Num) and stable Num
    }
} and

/* Skip past some spaces.  No token is output */
define lex_spaces() = {
    while not end_of_chars() and char_data() = " " do
	get_char()
} and

/* Lex operators */
define lex_operator() = exists Num: {
    if char_data() = "(" or
	char_data() = ")" or
	char_data() = "+"
    then {
	put_token(char_data());
	get_char()
    }
    else lexer_syntax_error("Unknown operator: "+char_data())
} and


/* Open the character and token streams */
define open_char_and_token_ports() = {
    {
	open_receiver_port(Char_receiver) and
	stable Token_sender
    };
    {
	open_sender_port(Token_sender) and
	stable Char_receiver
    }
} and
    

/* Test whether have another character */
define end_of_chars() = {
    end_of_receiver_port(Char_receiver)
} and

/* Get a new character */
define get_char() = {
    advance_receiver_port(Char_receiver) and
    stable Token_sender
} and

/* Get data from character port */
define char_data() = {
    receiver_port_data(Char_receiver)
} and

/* Output a token */
define put_token(token) = {
    advance_sender_port(Token_sender,token) and
    stable Char_receiver
} and

/* Leave the char and token streams idle */
define stable_ports()  = {
	stable Char_receiver and
	stable Token_sender
} and

/* Close the char and token streams */
define close_char_and_token_ports() = {
    {
	close_sender_port(Token_sender) and
	stable Char_receiver
    };
    {
	close_receiver_port(Char_receiver) and
	stable Token_sender
    }
} and
    
/* Terminate if syntax error detected */
define lexer_syntax_error(text) = {
    format("**Syntax error detected by lexer: %s\n",text) and
    false
}.


/*  Test routine */
define lexer_test(string_input) = {
    exists Char_stream,Token_stream,token_output,Sig_list: {
	stream(Char_stream) and stream(Token_stream)
	and format("Lexer started.  Input string = %t\n", string_input)
	and exec_subprocesses(Sig_list,3)
	and list_to_port_process(Sig_list[0],
	    string_input,sender_port(Char_stream))
	and lexer_process(Sig_list[1],receiver_port(Char_stream),
	    sender_port(Token_stream))
	and port_to_list_process(Sig_list[2],
	    receiver_port(Token_stream),token_output)
	and box format("Char_status=%8t  Token_status=%8t\n",
	    stream_status(Char_stream),stream_status(Token_stream))
	and fin format("Lexer finished.  Token list = %t\n", token_output)
    }
}.

/*  test data */
define lexer_test1 = "1".

define lexer_test2 = "(1)".

define lexer_test3 = "10 +2 ".

define lexer_test4 = " 10+ 32+  (57+1 ) ".

define lexer_test5 = " 10 / 32+  (57+1 ) ".

/*  end of stream-lexer file */

