/* -*- Mode: C -*-
 * Asynchronous expression parser using streams.  
 * 13 Jan 85.
 * Ben Moszkowski
 * Updated 23 Jan 85; 29 Jan 85; 30 Jan 85; 12 Feb 85; 19 Apr 85
 *
 * The process parser_process(Signal,Token_receiver,Polish_sender) takes the
 * token input from the port Token_receiver and parses it, producing
 * operators and operands on in reverse Polish on the output
 * port Polish_sender.
 *
 */

/* Uses stream routines. */
/* load(streams). */

/* Parser process */
define parser_process(Signal,Token_receiver,Polish_sender) = 
  process {
    {
	begin_process(Signal)
	and stable Token_receiver and stable Polish_sender
    };
    {
	parse_expr() and stable Signal
    };
    {
	finish_process(Signal)
	and stable Token_receiver and stable Polish_sender
    }
} where

/* Body of parser_process */
define parse_expr() = {
    open_token_and_polish_ports();
    parse_expression();
    if end_of_tokens() then close_token_and_polish_ports()
    else parser_syntax_error("Parser terminated before end of expression")
} and


/* Parse an expression */
define parse_expression() = {
    parse_primary();
    while not end_of_tokens() and current_token()="+" do {
	get_token();
	parse_primary();
	put_polish("+")
    }
} and



/* Parse a primary */
define parse_primary() = {
    if end_of_tokens() then
	parser_syntax_error("Premature end of expression")
    else if type(current_token())="integer" then {
	put_polish(current_token());
	get_token()
    }
    else if current_token()="(" then {
	{
	    get_token();
	    parse_expression();
	    if not end_of_tokens() and current_token()=")" then
		get_token()
	    else parser_syntax_error("Missing right parenthesis")
	}
    }
    else parser_syntax_error("Bad primary")
} and

/* Open the token and polish ports */
define open_token_and_polish_ports() = {
    {
	open_receiver_port(Token_receiver) and
	stable Polish_sender
    };
    {
	open_sender_port(Polish_sender) and
	stable Token_receiver
    }
} and
    
/* Test whether any more tokens */
define end_of_tokens() = {
    end_of_receiver_port(Token_receiver)
} and

/*  Value of current token */
define current_token() =  {
    receiver_port_data(Token_receiver)
} and

/* Get a new token */
define get_token() = {
    advance_receiver_port(Token_receiver) and
    stable Polish_sender
} and

/* Put a polish token */
define put_polish(polish) = {
    advance_sender_port(Polish_sender,polish) and
    stable Token_receiver
} and

/* Close the token and polish ports */
define close_token_and_polish_ports() = {
    {
	close_sender_port(Polish_sender) and
	stable Token_receiver
    };
    {
	close_receiver_port(Token_receiver) and
	stable Polish_sender
    }
} and
    

/* Terminate if syntax error detected */
define parser_syntax_error(text) = {
    format("**Syntax error detected by parser: %t\n",text)
    and false
}.



    
/* Test routine */
define parser_test(token_input) = {
    exists Token_stream,Polish_stream,polish_list,Sig_list: {
	stream(Token_stream) and stream(Polish_stream)
	and format("Parser started.  Token input list = %t\n", token_input)
	and exec_subprocesses(Sig_list,3)
	and list_to_port_process(Sig_list[0],
	    token_input,sender_port(Token_stream))
	and parser_process(Sig_list[1],
	    receiver_port(Token_stream),sender_port(Polish_stream))
	and port_to_list_process(Sig_list[2],
	    receiver_port(Polish_stream),polish_list)
	and box format("Token_status=%8t  Polish_status=%8t\n",
	    stream_status(Token_stream),stream_status(Polish_stream))
	and fin format("Parser finished.  Reverse Polish = %t\n", polish_list)
    }
}.

/* Test data */
define parser_test1 = [1].

define parser_test2 = ["(",1,")"].

define parser_test3 = [1,"+",2].

define parser_test4 = [1,"+",2,"+","(",3,"+",4,")"].

/*  End of stream-parser file */
