/* val.c: value checking routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: val.c,v 2.10 2001/09/10 15:17:39 acau Exp $
 */


#include "system.h"





/*****************\
*                 *
*  Type checking  *
*                 *
\*****************/

void type_check(e,type)
NODE1PTR e;
unsigned type;
{
    assert2 (e -> type == type,
            "found expression of type %s when expecting %s",
            print_name (e), Keyword[type]);
}

/********************************\
*                                *
*  Test whether val is defined   *
*                                *
\********************************/
int val_defined(e)
NODEPTR e;
{
    switch (e -> type) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case LAMBDA:
            return (T);
        case LOC:
            return (pointer_defined (locp(e)));
        case LIST:
            return (list_defined (listp(e)));
        default:
            error_break1 ("value cannot be of type %s", print_name(e));
            return (F);
    }
}

int pointer_defined(p)
LOCPTR p;
{
    return (mem_defined (p -> to_next ?
                next_loc (p -> ptr) :
                curr_loc (p -> ptr)));
}

/* Check whether list is defined. */
int list_defined(list)
LISTPTR list;
{
    register int     current_size;
    register unsigned i;
    register MEMORYPTR base_loc;
    MEMORYPTR size_loc;

    if (list_size_defined (list)) {
        base_loc = list_base (list);
        size_loc = list_size (list);
        current_size =  get_int (size_loc);
        for (i = 0; (int)i < current_size; i += 1) {
            if (!mem_defined (base_loc + offset (i)))
                return (F);
            if (!val_defined (get_mem (base_loc + offset (i))))
                return (F);
        }
        return (T);
    }
    else
        return (F);
}

int max_size(list)
LISTPTR list;
{
    MEMORYPTR size_loc;
    if (mem_defined (size_loc = list_size (list)) && is_static_mem (size_loc))
        return (get_int (size_loc));
    else
        return (list -> blk -> blk_size - 1);
}

/* Check whether current size of list is defined. */
int list_size_defined(list)
LISTPTR list;
{
    int     current_size;
    register MEMORYPTR size_loc = list_size (list);
    if (mem_defined (size_loc)) {
        current_size = get_int (size_loc);
        assert2d ((current_size >= 0) && (current_size <= max_size (list)),
                "current size of list is out of range [0,%d]: %d",
                max_size (list), current_size);
        return (T);
    }
    else
        return (F);
}



/********************************************************\
*                                                        *
*  Copying, comparing, assigning and outputting values.  *
*                                                        *
\********************************************************/

/* convert a value to a printable form */
void output_val(e)
NODEPTR e;
{
    switch (e -> type) {
        case NIL:
	    output_str ("nil");
	    break;
        case BOOL:
            output_str (e == True ? "true" : "false");
            break;
        case INT:
            output_int (intp (e) -> ival, 0);
            break;
        case STRING:
            output_char ('"');
            output_str (stringp (e) -> str);
            output_char ('"');
            break;
        case LIST:
            output_list (listp(e));
            break;
        case LOC:
            output_char ('@');
            output_int ((int)cell_no (locp (e) -> ptr), 0);
            break;
        case LAMBDA:
            output_lambda (e);
            break;   
        default:
            error_break1 ("expression of type %s cannot be output",
                    print_name (e));
    }
}

void output_list(list)
LISTPTR list;
{
    int     size;
    unsigned i = 0;
    MEMORYPTR loc,
            base_loc;

    base_loc = list_base (list);
    size = mem_defined (loc = list_size (list)) ? get_int (loc) : max_size (list);
    output_char ('[');
    while ((int)i < size) {
        if (mem_defined (loc = base_loc + offset (i)))
            output_val (get_mem (loc));
        else
            output_char ('?');
        if ((int)++i < size)
            output_char (',');
    }
    output_char (']');
}


/* Return a copy of a value. Atoms need no copying. */
NODEPTR copy_val(val)
NODEPTR val;
{
    switch (val -> type) {
        case NIL:
        case BOOL:
        case INT:
        case LAMBDA:
        case STRING:
        case LOC:
        case STRUCT:
            return (val);
        case LIST:
            return (nodep (copy_list (listp(val))));
        default:
            error_break1 ("cannot copy value of type %s", print_name(val));
            return (F);
    }
}

/* Make a copy of a list. */
/* The copy is defined in the CURRENT state. */
LISTPTR copy_list(list)
LISTPTR list;
{
    MEMORYPTR size_loc = list_size (list);
    MEMORYPTR base_loc = list_base (list);
    MEMORYPTR new_base;
    LISTPTR list1;
    int i, size;

/*printf("***Copying List\n");*/
    assert0 (list_defined (list),
            "attempt to copy an incompletely defined list");
    list1 = new_list (STATIC, STATIC, intp(get_mem (size_loc)));
/*    if (is_static_mem (base_loc) && is_static_mem (size_loc)) {*/
/*printf("***SAVED\n");*/
/*        return (list);*/
/*	}*/
    list = list1;
    new_base = curr_loc (list_base (list));
    size = get_int (size_loc);
    for (i = 0; i < size; i += 1)
        put_mem (new_base + offset (i),
                copy_val (get_mem (base_loc + offset (i))));
    return (list);
}

int string_size(s)
STRINGPTR s;
{
    return (strlen (s -> str));
}

NODEPTR substring(s,i,j)
STRINGPTR s;
int i,j;
{
    return (string_node (substr (s -> str, i, j)));
}

NODEPTR append_string (s1, s2)
STRINGPTR s1, s2;
{
    return (string_node (strapnd (s1 -> str, s2 -> str)));
}

/* Return the result of appending list2 to the end of list1. */
/* The new list is defined in the CURRENT state. */
LISTPTR append_list(list1,list2)
LISTPTR list1,list2;
{
    MEMORYPTR     new_base,
            base_loc1,
            base_loc2;
    int     size1,
            size2;
    unsigned i;

    base_loc1 = list_base (list1);
    size1 = get_int (list_size (list1));
    base_loc2 = list_base (list2);
    size2 = get_int (list_size (list2));
    list1 = new_list (STATIC, STATIC, intp(int_node (size1 + size2)));
    new_base = curr_loc (list1 -> base);
    for (i = 0; (int)i < size1; i += 1)
        put_mem (new_base + offset (i),
                copy_val (get_mem (base_loc1 + offset (i))));
    new_base += offset (i);
    for (i = 0; (int)i < size2; i += 1)
        put_mem (new_base + offset (i),
                copy_val (get_mem (base_loc2 + offset (i))));
    return (list1);
}

/* Create a new list. */
LISTPTR new_list(cell_type,size_type,n)
int cell_type,size_type;
INTPTR n;
{
    BLOCKPTR blk = mem_block (cell_type, (unsigned)n -> ival + 1);
    MEMORYPTR size_loc = blk -> blk_base,
        base_loc = size_loc + offset (1);

    new_loc (size_loc, size_type);
    if (size_type == STATIC)
        put_mem (size_loc, nodep(n));
    return (listp (list_node (blk, size_loc, base_loc)));
}

/* Assign a value to a location */
void assign_val (loc,val)
NODEPTR val;
register MEMORYPTR loc;
{
    switch (val -> type) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case LAMBDA:
        case LOC:
            if (mem_defined (loc)) {
                if (RedundancyFlag || !equal_val (val, get_mem (loc))) {
		    if (loc == TempLoc)
		        error_break0 ("attempt to redefine constant");
		    else
		        error_break0 ("attempt to overwrite variable");
		}
            }
            else
                put_mem (loc, val);
            break;
        case STRUCT:
            assign_struct (loc, listp(val -> c1));
            break;
        case LIST:
            assign_list (loc, listp(val));
            break;
        default:
            error_break1 ("cannot assign value of type %s", print_name (val));
    }
}

void assign_struct(loc,struct1)
MEMORYPTR loc;
LISTPTR struct1;
{
    if (!mem_defined (loc))
        put_mem (loc, nodep(struct1));
    else {
	register int size2;
        register MEMORYPTR loc1;
        register MEMORYPTR loc2;
        LISTPTR struct2 = get_list (loc);
        type_check (nodep(struct2), LIST);
        struct1 = put_list (loc, nodep(struct1)); /* possibly adjusted */
        assert0 (max_size (struct1) >= max_size (struct2),
	         "attempt to redefine max. size of list");
        if (mem_defined (struct2 -> size)) {
	    size2 = get_int (struct2 -> size);
	    if (struct1 -> size != struct2 -> size) {
	        put_mem (struct1 -> size, get_mem (struct2 -> size));
                make_indirect_mem (struct2 -> size, struct1 -> size);
	    }
            loc1 = struct1 -> base;
            loc2 = struct2 -> base;
	    if (loc1 != loc2) {
                while (size2-- > 0) {
	            if (mem_defined (loc2))
                        assign_val (loc1, get_mem (loc2));
                    make_indirect_mem (loc2, loc1);
                    loc1 += offset(1);
                    loc2 += offset(1);
		}
	    }
	}
    }
}

/* copy one list into another. */
void assign_list(loc,list)
MEMORYPTR loc;
LISTPTR list;
{
    int     current_size1,
            current_size2;
    register MEMORYPTR loc1;
    register MEMORYPTR loc2;
    MEMORYPTR size_loc1;
    MEMORYPTR size_loc2;
    LISTPTR list1;

    loc2 = list_base (list);
    size_loc2 = list_size (list);
    current_size2 = get_int (size_loc2);
    if (mem_defined (loc)) {
        assert0 (is_list (list1 = get_list (loc)),
                "attempt to overwrite non-list with list");
        assert2d (max_size (list1) >= current_size2,
                "attempt to assign list of size %d to one of max size %d",
                current_size2, max_size (list1));
    }
    else {
        list1 = new_list (is_static_mem (loc) ? STATIC : STATE,
                STATIC, intp(get_mem (size_loc2)));
        list1 = put_list (loc, nodep(list1));/* possibly adjusted. */
    }
    loc1 = list_base (list1);
    size_loc1 = list_size (list1);

    if (!mem_defined (size_loc1)) {
        current_size1 = current_size2;
        put_mem (size_loc1, int_node (current_size1));
    }
    else {
        assert2d ((current_size1 = get_int (size_loc1)) == current_size2,
                "attempt to assign list of size %d to one of size %d",
                current_size2, current_size1);
    }
    while (current_size2-- > 0) {
        if (mem_defined(loc2))
            assign_val (loc1, get_mem (loc2));
	loc1 += offset (1);
	loc2 += offset (1);
    }
}

/* Test whether two values are equal. */
int equal_val(e1,e2)
NODEPTR e1,e2;
{
    if (e1 -> type != e2 -> type)
        return (F);
    return (compare_val (intp(e1), intp(e2)) == 0);
}


/* Compare two values. Return 0 if e1=e2, >0 if e1>e2, <0 if e1<e2. */
int compare_val(e1,e2)
INTPTR e1,e2;
{
    unsigned type;

    if ((type = e1 -> type) != e2 -> type)
        return (-1);
    switch (type) {
        case NIL:
	    return (0);
        case BOOL:
        case INT:
            return (e1 -> ival - e2 -> ival);
        case LAMBDA:
            return (e1 - e2);
        case LOC:
            return (compare_ptrs (locp(e1), locp(e2)));
        case STRING:
            return (compare_strings (stringp(e1), stringp(e2)));
        case LIST:
            return (compare_lists (listp(e1), listp(e2)));
        default:
            error_break1 ("cannot compare values of type %s", print_name (nodep(e1)));
            return (F);
    }
}

int compare_strings(e1,e2)
STRINGPTR e1,e2;
{
    return (strcmp (e1 -> str, e2 -> str));
}

int compare_ptrs(e1,e2)
LOCPTR e1,e2;
{
            return (e1 -> to_next == e2 -> to_next ?
                    e1 -> ptr - e2 -> ptr :
                    e1 -> to_next - e2 -> to_next);
}

int compare_lists(list1,list2)
LISTPTR list1,list2;
{
    int     size1,
            size2,
            answer;
    unsigned i;
    MEMORYPTR base_loc1,
            base_loc2;

    assert0 (list_defined (list1) && list_defined (list2),
            "attempt to compare an undefined list");
    base_loc1 = list_base (list1);
    base_loc2 = list_base (list2);
    size1 = get_int (list_size (list1));
    size2 = get_int (list_size (list2));
    if (size1 != size2)
        return (size1 - size2);
    for (i = 0; (int)i < size1; i += 1) {
        if ((answer = compare_val (intp(get_mem (base_loc1 + offset (i))),
                    intp(get_mem (base_loc2 + offset (i))))) != 0)
            return (answer);
    }
    return (0);
}

